// Copyright (C) 2015 Ryan Curtin
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

#if defined(ARMA_USE_SUPERLU)

namespace superlu
  {


  template<typename eT>
  inline
  void
  convert_to_supermatrix(const SpMat<eT>& A, SuperMatrix& out)
    {
    // We store in column-major CSC.
    out.Stype = SLU_NC;

    if (is_float<eT>::value == true)
      {
      out.Dtype = SLU_S;
      }
    else
    if (is_double<eT>::value == true)
      {
      out.Dtype = SLU_D;
      }
    else
    if (is_supported_complex_float<eT>::value == true)
      {
      out.Dtype = SLU_C;
      }
    else
    if (is_supported_complex_double<eT>::value == true)
      {
      out.Dtype = SLU_Z;
      }

    out.Mtype = SLU_GE; // Just a general matrix.  We don't know more now.

    // We have to actually create the object which stores the data.
    // This gets cleaned by destroy_supermatrix().
    // We have to use SuperLU's stupid memory allocation routines since they are
    // not guaranteed to be new and delete.  We can't overload SUPERLU_MALLOC()
    // to be new and delete either, because it won't handle the array case.
    NCformat* nc = (NCformat*) SUPERLU_MALLOC(sizeof(NCformat));

    nc->nnz = A.n_nonzero;
    nc->nzval = (void*) SUPERLU_MALLOC(sizeof(eT) * nc->nnz);
    nc->colptr = (int_t*) SUPERLU_MALLOC(sizeof(int_t) * (A.n_cols + 1));
    nc->rowind = (int_t*) SUPERLU_MALLOC(sizeof(int_t) * nc->nnz);

    // Fill the matrix.
    arrayops::copy((eT*) nc->nzval, A.values, A.n_nonzero);
    // These have to be copied by hand, because the types may differ.
    for (uword i = 0; i <= A.n_cols; ++i)
      nc->colptr[i] = (int_t) A.col_ptrs[i];
    for (uword i = 0; i < A.n_nonzero; ++i)
      nc->rowind[i] = (int_t) A.row_indices[i];

    out.nrow = A.n_rows;
    out.ncol = A.n_cols;
    out.Store = (void*) nc;
    }



  template<typename eT>
  inline
  void
  convert_to_supermatrix(const Mat<eT>& A, SuperMatrix& out)
    {
    // This is being stored as a dense matrix.
    out.Stype = SLU_DN;

    if (is_float<eT>::value == true)
      {
      out.Dtype = SLU_S;
      }
    else
    if (is_double<eT>::value == true)
      {
      out.Dtype = SLU_D;
      }
    else
    if (is_supported_complex_float<eT>::value == true)
      {
      out.Dtype = SLU_C;
      }
    else
    if (is_supported_complex_double<eT>::value == true)
      {
      out.Dtype = SLU_Z;
      }

    out.Mtype = SLU_GE;

    // We have to create the object that stores the data.
    DNformat* dn = (DNformat*) SUPERLU_MALLOC(sizeof(DNformat));

    dn->lda = A.n_rows;
    dn->nzval = (void*) A.memptr();

    out.nrow = A.n_rows;
    out.ncol = A.n_cols;
    out.Store = (void*) dn;
    }



  inline
  void
  destroy_supermatrix(SuperMatrix& out)
    {
    // Clean up.
    if (out.Stype == SLU_NC)
      {
      NCformat* nc = (NCformat*) out.Store;

      SUPERLU_FREE(nc->nzval);
      SUPERLU_FREE(nc->colptr);
      SUPERLU_FREE(nc->rowind);
      SUPERLU_FREE(nc);
      }
    else if (out.Stype == SLU_DN)
      {
      DNformat* dn = (DNformat*) out.Store;

      SUPERLU_FREE(dn);
      }
    else
      {
      // Uh, crap.  (but this should never happen unless the user does something
      // catastrophically stupid.)
      }
    }

  }

#endif
